#ifndef OBJECTRELATIONALACCESS_OBJECTRELATIONALDEFAULTMAPPINGRULES_H
#define OBJECTRELATIONALACCESS_OBJECTRELATIONALDEFAULTMAPPINGRULES_H

#include <string>
#include <locale>
#include <algorithm>

namespace pool {

  class ObjectRelationalDefaultMappingRules {
    
    public:

      /// Returns the name of the index for the identity of an object mapped to a table
      static std::string indexNameForIdentity( const std::string& tableName );
      /// Returns the name of the foreign key to the identity of the parrent table
      static std::string fkNameForIdentity( const std::string& tableName );

      static std::string newMappingVersionForClass( const std::string& className, int iteration );

      static std::string classVersionPropertyNameInDictionary();

      static std::string defaultClassVersion(const std::string& className);

    public:
      /// Returns the default name for a table given a class name
      static std::string tableNameForClass( const std::string& className );
      static std::string tableNameForClass( const std::string& className,const std::string& scope );
      static std::string columnNameForId();
      static std::string columnNameForPosition();
      static std::string columnNameForVariable( const std::string& variableName, const std::string& scope );
      static std::string columnNameForOID_1( const std::string& variableName, const std::string& scope );
      static std::string columnNameForOID_2( const std::string& variableName, const std::string& scope );

    private:

      struct ToUpper {
          ToUpper(const std::locale& l):loc(l) {}
          char operator() (char c) const  { return std::toupper(c,loc); }
          std::locale const& loc;
      };
  };
}

#endif

