#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingRules.h"
#include <sstream>
#include <vector>

std::string
pool::ObjectRelationalDefaultMappingRules::indexNameForIdentity( const std::string& tableName )
{
  return tableName + "_ID_IDX";
}

std::string
pool::ObjectRelationalDefaultMappingRules::fkNameForIdentity( const std::string& tableName )
{
  return tableName + "_ID_FK";
}

std::string
pool::ObjectRelationalDefaultMappingRules::newMappingVersionForClass( const std::string& className, int iteration )
{
  std::ostringstream os;
  os << className;
  os << "_";
  if ( iteration > 0 ) {
    if ( iteration < 10 ) os << "0";
    if ( iteration < 100 ) os << "0";
    os << iteration;
  } else {
    os << "000";
  }
  
  return os.str();
}

std::string
pool::ObjectRelationalDefaultMappingRules::classVersionPropertyNameInDictionary()
{
  static std::string s_propertyName("class_version");
  return s_propertyName;

}

std::string
pool::ObjectRelationalDefaultMappingRules::defaultClassVersion(const std::string& className)
{
  std::string classVersion(className);
  classVersion.append("_default");
  return classVersion;
}

std::string
pool::ObjectRelationalDefaultMappingRules::tableNameForClass( const std::string& className )
{
  std::string myname(className);
  //turn className into uppercase
  ObjectRelationalDefaultMappingRules::ToUpper up(std::locale::classic());
  std::transform(myname.begin(), myname.end(),myname.begin(),up);
  std::ostringstream result;
  char prev = 'a';
  for ( std::string::size_type i = 0; i < myname.size(); ++i ) {
    char c = myname[i];
    if ( c == ':' || c == '<' || c == '>' || c == ',' ) c = '_';
    if( c != ' ' ) {
      if(c!='_' || prev!='_'){
        if(c!='_' || i!=0) {
          result << c;
          prev = c;
        }
      }
    }
  }
  return result.str();
}


std::string
pool::ObjectRelationalDefaultMappingRules::tableNameForClass( const std::string& className,
                                                              const std::string& parentTableName )
{
  std::ostringstream result;
  if ( ! parentTableName.empty() ) result << parentTableName << "_";
  result << pool::ObjectRelationalDefaultMappingRules::tableNameForClass( className );
  return result.str();
}


std::string
pool::ObjectRelationalDefaultMappingRules::columnNameForId()
{
  return std::string("ID");
}


std::string
pool::ObjectRelationalDefaultMappingRules::columnNameForPosition()
{
  return std::string("POS");
}

std::string
pool::ObjectRelationalDefaultMappingRules::columnNameForOID_1( const std::string& variableName, const std::string& scope )
{
  std::string tempOutput = "OID_1_" + variableName;
  return columnNameForVariable( tempOutput, scope );
}

std::string
pool::ObjectRelationalDefaultMappingRules::columnNameForOID_2( const std::string& variableName, const std::string& scope )
{
  std::string tempOutput = "OID_2_" + variableName;
  return columnNameForVariable( tempOutput, scope );
}

std::string
pool::ObjectRelationalDefaultMappingRules::columnNameForVariable( const std::string& variableName,
                                                                  const std::string& scopeName )
{
  std::vector<std::string> scopes;
  std::string mscope(scopeName);
  std::string delim("::");
  std::string::size_type begIdx,endIdx;
  std::string elem;
  while( (begIdx=mscope.find_first_not_of(delim))!= mscope.npos ){
    endIdx=mscope.find_first_of(delim);
    if(endIdx != mscope.npos ){
      elem=mscope.substr(begIdx,endIdx-begIdx);
      mscope=mscope.substr(endIdx+delim.length());
      scopes.push_back(elem);
    }else{
      elem=mscope;
      scopes.push_back(elem);
      break;
    }
  }
  std::ostringstream result;
  std::vector<std::string>::iterator beg=scopes.begin();
  for( std::vector<std::string>::iterator it=beg; it!=scopes.end(); ++it ){
    if( it != beg){
      std::string myname(*it);
      ObjectRelationalDefaultMappingRules::ToUpper up(std::locale::classic());
      std::transform(myname.begin(), myname.end(),myname.begin(),up);
      for(std::string::size_type pos=myname.find("::", 0);
          pos!=std::string::npos;
          pos = myname.find("::",pos+1)){
        myname.replace(pos,2,"_");
      }
      for ( std::string::size_type i = 0; i < myname.size(); ++i ) {
        char c = myname[i];
        if ( c == ':' || c == '<' || c == '>' || c == ',' || c == '[' ) c = '_';
        if( c != ' ' &&  c!= ']' ) result << c;
      }
      result << "_";
    }
  }

  
  std::string myname(variableName);
  //turn className into uppercase
  ObjectRelationalDefaultMappingRules::ToUpper up(std::locale::classic());
  std::transform(myname.begin(), myname.end(),myname.begin(),up);
  
  for(std::string::size_type pos=myname.find("::", 0);
      pos!=std::string::npos;
      pos = myname.find("::",pos+1)){
    myname.replace(pos,2,"_");
  }
  for ( std::string::size_type i = 0; i < myname.size(); ++i ) {
    char c = myname[i];
    if ( c == ':' || c == '<' || c == '>' ||  c == ',' || c == '[' ) c = '_';
    if( c != ' ' &&  c!= ']' ) {
      if( c != '_' || i<(myname.size()-1)) result << c;
    }
  }

  return result.str();
}
