#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include <sstream>

std::string
pool::ObjectRelationalMappingElement::objectMappingElementType()
{
  static std::string s_objectMappingElementType = "Object";
  return s_objectMappingElementType;
}

std::string
pool::ObjectRelationalMappingElement::primitiveMappingElementType()
{
  static std::string s_primitiveMappingElementType = "Primitive";
  return s_primitiveMappingElementType;
}

std::string
pool::ObjectRelationalMappingElement::arrayMappingElementType()
{
  static std::string s_arrayMappingElementType = "Array";
  return s_arrayMappingElementType;
}

std::string
pool::ObjectRelationalMappingElement::CArrayMappingElementType()
{
  static std::string s_CArrayMappingElementType = "CArray";
  return s_CArrayMappingElementType;
}

std::string
pool::ObjectRelationalMappingElement::inlineCArrayMappingElementType()
{
  static std::string s_inlineCArrayMappingElementType = "InlineCArray";
  return s_inlineCArrayMappingElementType;
}

std::string
pool::ObjectRelationalMappingElement::poolReferenceMappingElementType()
{
  static std::string s_poolReferenceMappingElementType = "PoolReference";
  return s_poolReferenceMappingElementType;
}

std::string
pool::ObjectRelationalMappingElement::poolPointerMappingElementType()
{
  static std::string s_poolPointerMappingElementType = "PoolPointer";
  return s_poolPointerMappingElementType;
}

std::string
pool::ObjectRelationalMappingElement::poolArrayMappingElementType()
{
  static std::string s_poolArrayMappingElementType = "PoolArray";
  return s_poolArrayMappingElementType;
}

std::string
pool::ObjectRelationalMappingElement::pointerMappingElementType()
{
  static std::string s_pointerMappingElementType = "Pointer";
  return s_pointerMappingElementType;
}

std::string
pool::ObjectRelationalMappingElement::referenceMappingElementType()
{
  static std::string s_referenceMappingElementType = "Reference";
  return s_referenceMappingElementType;
}

std::string
pool::ObjectRelationalMappingElement::blobMappingElementType()
{
  static std::string s_blobMappingElementType = "Blob";
  return s_blobMappingElementType;
}

bool
pool::ObjectRelationalMappingElement::isValidMappingElementType( const std::string& elementType )
{
  return ( elementType == objectMappingElementType() ||
           elementType == primitiveMappingElementType() ||
           elementType == arrayMappingElementType() ||
           elementType == CArrayMappingElementType() ||
           elementType == inlineCArrayMappingElementType() ||
           elementType == poolReferenceMappingElementType() ||
           elementType == poolPointerMappingElementType() ||
           elementType == poolArrayMappingElementType() ||
           elementType == pointerMappingElementType() ||
           elementType == referenceMappingElementType() ||
           elementType == blobMappingElementType() );
}

std::string
pool::ObjectRelationalMappingElement::elementTypeAsString( pool::ObjectRelationalMappingElement::ElementType elementType )
{
  switch ( elementType ) {
  case pool::ObjectRelationalMappingElement::Object :
    return  objectMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::Primitive :
    return  primitiveMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::Array :
    return  arrayMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::CArray :
    return  CArrayMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::InlineCArray :
    return  inlineCArrayMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::PoolReference :
    return  poolReferenceMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::PoolPointer :
    return  poolPointerMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::PoolArray :
    return  poolArrayMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::Reference :
    return  referenceMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::Blob :
    return  blobMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::Pointer :
    return  pointerMappingElementType();
    break;
  case pool::ObjectRelationalMappingElement::Undefined :
    // This should never appear
    break;
  };

  throw pool::ObjectRelationalException( "Undefined mapping element type",
                                         "ObjectRelationalMappingElement::elementTypeAsString" );
  return "";
}


pool::ObjectRelationalMappingElement::ElementType
pool::ObjectRelationalMappingElement::elementTypeFromString( const std::string& elementType )
{
  // Check here the element type
  if ( !  isValidMappingElementType( elementType ) ) {
    throw pool::ObjectRelationalException( "\"" + elementType + "\" is not a supported mapping element type",
                                           "ObjectRelationalMappingElement::elementTypeFromString" );
  }

  pool::ObjectRelationalMappingElement::ElementType result = pool::ObjectRelationalMappingElement::Undefined;

  if ( elementType == objectMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::Object;
  }
  if ( elementType == primitiveMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::Primitive;
  }
  else if ( elementType == arrayMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::Array;
  }
  else if ( elementType == CArrayMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::CArray;
  }
  else if ( elementType == inlineCArrayMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::InlineCArray;
  }
  else if ( elementType == pointerMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::Pointer;
  }
  else if ( elementType == referenceMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::Reference;
  }
  else if ( elementType == poolReferenceMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::PoolReference;
  }
  else if ( elementType == poolPointerMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::PoolPointer;
  }
  else if ( elementType == poolArrayMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::PoolArray;
  }
  else if ( elementType == blobMappingElementType() ) {
    return pool::ObjectRelationalMappingElement::Blob;
  }

  return result;
}

std::string
pool::ObjectRelationalMappingElement::variableNameForArray( const std::string& arrayVariable, unsigned int index ){
  std::ostringstream arrayElementLabel;
  arrayElementLabel << arrayVariable << "[" << index << "]";
  return arrayElementLabel.str();
}

std::string
pool::ObjectRelationalMappingElement::scopedVariableName( const std::string& variableName, const std::string& scope){
  std::string scopedName("");
  if(!scope.empty()) scopedName.append(scope).append("::");
  scopedName.append(variableName);
  return scopedName;
}

pool::ObjectRelationalMappingElement::ObjectRelationalMappingElement( const std::string& elementType,
                                                                      const std::string& variableName,
                                                                      const std::string& variableType,
                                                                      const std::string& tableName ):
  m_elementType( pool::ObjectRelationalMappingElement::Undefined ),
  m_scopeName( "" ),
  m_variableName( variableName ),
  m_variableType( variableType ),
  m_tableName( tableName ),
  m_columnNames(),
  m_subElements(),
  m_scopeForColumnName(""),
  m_fullScope( variableName )
{
  // Check here the element type
  if ( !  isValidMappingElementType( elementType ) ) {
    throw pool::ObjectRelationalException( "\"" + elementType + "\" is not a supported mapping element type",
                                           "ObjectRelationalMappingElement" );
  }
  m_elementType = pool::ObjectRelationalMappingElement::elementTypeFromString( elementType );
}

std::set<std::string>
pool::ObjectRelationalMappingElement::involvedTables() const{
  std::set<std::string> involvedTables;
  involvedTables.insert(tableName());
  const_iterator iel = begin();
  for(const_iterator iel=begin();
      iel!=end();++iel){
    std::set<std::string> elementTables = iel->second.involvedTables();
    for(std::set<std::string>::const_iterator iT=elementTables.begin();
        iT!=elementTables.end();++iT){
      involvedTables.insert(*iT);
    }
  }
  return involvedTables;
}

pool::ObjectRelationalMappingElement&
pool::ObjectRelationalMappingElement::appendSubElement( const std::string& elementType,
                                                        const std::string& variableName,
                                                        const std::string& variableType,
                                                        const std::string& tableName )
{
  // Check whether the current element type supports other subelements
  if ( ! ( m_elementType == pool::ObjectRelationalMappingElement::Object ||
           m_elementType == pool::ObjectRelationalMappingElement::Array ||
           m_elementType == pool::ObjectRelationalMappingElement::CArray ||
           m_elementType == pool::ObjectRelationalMappingElement::InlineCArray ||
           m_elementType == pool::ObjectRelationalMappingElement::PoolReference ||
           m_elementType == pool::ObjectRelationalMappingElement::PoolPointer ||
           m_elementType == pool::ObjectRelationalMappingElement::PoolArray ) ) {
    throw pool::ObjectRelationalException( "Attempted to insert a sub-element under an element of type \"" +
                                           pool::ObjectRelationalMappingElement::elementTypeAsString( m_elementType ) + "\"",
                                           "ObjectRelationalMappingElement::appendSubElement" );
  }
  pool::ObjectRelationalMappingElement& newElement = m_subElements.insert( std::make_pair( variableName,
                                                                                           pool::ObjectRelationalMappingElement( elementType,
                                                                                                                                 variableName,
                                                                                                                                 variableType,
                                                                                                                                 tableName ) ) ).first->second;
  // Determine here the scope of the new element
  if ( m_scopeName.empty() ) {
    newElement.m_scopeName = m_variableName;
    newElement.m_scopeForColumnName = m_variableName; 
  }
  else {
    newElement.m_scopeName = m_scopeName + "::" + m_variableName;
    if( m_elementType == pool::ObjectRelationalMappingElement::InlineCArray){
      newElement.m_scopeForColumnName = m_scopeForColumnName;
    } else {
      newElement.m_scopeForColumnName = m_scopeForColumnName + "::" + m_variableName;
    }
  }
  newElement.m_fullScope = m_fullScope + "::" + variableName;
  return newElement;
}

void
pool::ObjectRelationalMappingElement::appendSubElementTree( const ObjectRelationalMappingElement& element ){
  if ( ! ( m_elementType == pool::ObjectRelationalMappingElement::Object ||
           m_elementType == pool::ObjectRelationalMappingElement::Array ||
           m_elementType == pool::ObjectRelationalMappingElement::CArray ||
           m_elementType == pool::ObjectRelationalMappingElement::InlineCArray ||
           m_elementType == pool::ObjectRelationalMappingElement::PoolReference ||
           m_elementType == pool::ObjectRelationalMappingElement::PoolPointer ||
           m_elementType == pool::ObjectRelationalMappingElement::PoolArray ) ) {
    throw pool::ObjectRelationalException( "Attempted to insert a sub-element under an element of type \"" +
                                           pool::ObjectRelationalMappingElement::elementTypeAsString( m_elementType ) + "\"",
                                           "ObjectRelationalMappingElement::appendSubElement" );
  }
  m_subElements.insert( std::make_pair(element.variableName(),element));
}

void
pool::ObjectRelationalMappingElement::alterType( const std::string& elementType )
{
  // Check here the element type
  if ( !  isValidMappingElementType( elementType ) ) {
    throw pool::ObjectRelationalException( "\"" + elementType + "\" is not a supported mapping element type",
                                           "ObjectRelationalMappingElement::alterType");
  }
  pool::ObjectRelationalMappingElement::ElementType elementTypeCode = pool::ObjectRelationalMappingElement::elementTypeFromString( elementType );
  if(m_elementType != elementTypeCode){
    m_elementType = elementTypeCode;      
    // clear sub elements when no supported by the new type specified
    if ( ! ( m_elementType == pool::ObjectRelationalMappingElement::Object ||
             m_elementType == pool::ObjectRelationalMappingElement::Array ||
             m_elementType == pool::ObjectRelationalMappingElement::CArray ||
             m_elementType == pool::ObjectRelationalMappingElement::PoolReference ||
             m_elementType == pool::ObjectRelationalMappingElement::PoolPointer ||
             m_elementType == pool::ObjectRelationalMappingElement::PoolArray ) ) {
      m_subElements.clear();
    }
  }
}

void
pool::ObjectRelationalMappingElement::alterTableName( const std::string& tableName )
{
  if ( tableName == m_tableName ) return;

  for ( pool::ObjectRelationalMappingElement::iterator iElement = this->begin();
        iElement != this->end(); ++iElement ) {
    pool::ObjectRelationalMappingElement& subElement = iElement->second;
    if ( subElement.elementType() != pool::ObjectRelationalMappingElement::Array &&
         subElement.elementType() != pool::ObjectRelationalMappingElement::CArray &&
         subElement.elementType() != pool::ObjectRelationalMappingElement::PoolPointer &&
         subElement.elementType() != pool::ObjectRelationalMappingElement::PoolArray ) {
      subElement.alterTableName( tableName );
    }
  }
  m_tableName = tableName;
}


void
pool::ObjectRelationalMappingElement::setColumnNames( const std::vector< std::string >& columns )
{
  m_columnNames = columns;
  if ( m_subElements.empty() ) return;

  std::vector< std::string > columnsToPropagate = m_columnNames;
  if ( this->elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    // Remove the OID fields
    columnsToPropagate.pop_back();
    columnsToPropagate.pop_back();
  }

  for ( pool::ObjectRelationalMappingElement::iterator iElement = this->begin();
        iElement != this->end(); ++iElement ) {
    pool::ObjectRelationalMappingElement& subElement = iElement->second;

    if ( subElement.elementType() == pool::ObjectRelationalMappingElement::Object &&
         subElement.tableName() == m_tableName ) {
      subElement.setColumnNames( columnsToPropagate );
    }
    else if ( subElement.elementType() == pool::ObjectRelationalMappingElement::PoolReference &&
              subElement.tableName() == m_tableName ) {
      std::vector< std::string > columns = subElement.columnNames();
      for ( size_t i = 0; i < columns.size() - 2; ++i ) columns[i] = columnsToPropagate[i];
      subElement.setColumnNames( columns );
    }
  }
}

void pool::ObjectRelationalMappingElement::override(const ObjectRelationalMappingElement& source)
{
  if(variableType()==source.variableType() || elementType()==source.elementType())
  {
    alterType(ObjectRelationalMappingElement::elementTypeAsString(source.elementType()));
    alterTableName(source.tableName());
    setColumnNames(source.columnNames());
    for(iterator iel=begin();
        iel!=end();++iel){
      const_iterator iTarg = source.find(iel->first);
      if(iTarg!=source.end()){
        iel->second.override(iTarg->second);
      }
    }
  }  

}

