#ifndef RELATIONALSTORAGESERVICE_MAPPING_MANAGER_H
#define RELATIONALSTORAGESERVICE_MAPPING_MANAGER_H

#include "Reflex/Type.h"
#include <string>
#include <map>
#include "RVersion.h"

namespace coral {

  class ISchema;
  
}

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
  typedef ROOT::Reflex::Type TypeH;
#else
  typedef Reflex::Type TypeH;
#endif

namespace pool {
  
  class ObjectRelationalMappingPersistency;

  namespace RelationalStorageService {

    class MappingInformation;

    /**
       @class MappingManager MappingManager.h
       Utility class to manage the object-relational mappings for the C++ classes.
    */

    class MappingManager {
    public:
      /// Constructor
      explicit MappingManager( coral::ISchema& schema );

      /// Destructor
      ~MappingManager();

      /// Retrieves a valid mapping for a given class.
      const MappingInformation& mappingForNewHomogeneousContainer( const std::string& containerName, const TypeH& dictionary );

      /// Retrieves a valid mapping for a given class to be used by an homogeneous container
      const MappingInformation& existingMappingForHomogeneousContainer( const std::string& containerName,
                                                                        const std::string& className,
                                                                        const std::string& nominalVersion );

      /// Retrieves a valid mapping for a given class to be used by a polymorphic container.
      const MappingInformation& mappingForPolymorphicContainer( const std::string& containerName, const TypeH& dictionary );

      /// Retrieves a valid mapping for a given class to be used by a polymorphic container
      const MappingInformation& existingMappingForPolymorphicContainer( const std::string& containerName,
                                                                        const TypeH& dictionary,
                                                                        const std::string& nominalVersion );

      /// Clears the internal state
      void clear();
      public:

      static std::string versionOfClass( const TypeH& dictionary );

      static std::string mappingVersionLookupKey( const std::string classVersion, const std::string containerName );

    private:
      /// Checks if a table is empty
      bool isTableEmpty( const std::string& tableName );

    private:
      /// The schema in use
      coral::ISchema& m_schema;
      /// The mapping persistency service
      pool::ObjectRelationalMappingPersistency* m_persistency;
      std::map< std::string, std::string > m_mappingVersionsForContainers;
      /// The map of mapping information objects for homogeneous containers
      std::map< const TypeH, std::map< std::string, MappingInformation* > > m_classMappingsForHomogeneousContainers;
      /// The map of mapping information objects for polymorphic containers.
      std::map< const TypeH, std::map< std::string, MappingInformation* > > m_classMappingsForPolymorphicContainers;
    };

  }

}

#endif
