#ifndef RELATIONALSTORAGESERVICE_OBJECTREADER_H
#define RELATIONALSTORAGESERVICE_OBJECTREADER_H

#include "DbIOHandler.h"
#include "Reflex/Type.h"
#include <vector>
#include <set>

#include "RVersion.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
 typedef ROOT::Reflex::Type TypeH;
using namespace ROOT;
#else
 typedef Reflex::Type TypeH;
#endif

namespace coral {

  class ISchema;
  class IQuery;
  class ICursor;
}

namespace pool {

  class ObjectRelationalMappingElement;
  class DataCallBack;
  class IPtrLoader;

  namespace RelationalStorageService {

    class ObjectRelationalReadDataBuffer;

    class ObjectReader : virtual public DbIOHandler {
    public:
      /// Constructor
      ObjectReader( coral::ISchema& schema,
                    const pool::ObjectRelationalMappingElement& mapping,
                    pool::DbDatabase* databaseHandle );

      ObjectReader( coral::ISchema& schema,
                    const TypeH& dictionary,
                    const pool::ObjectRelationalMappingElement& mapping,
                    pool::DbDatabase* databaseHandle );

      /// Destructor
      ~ObjectReader();

      /// Reads an object
      bool readObject( void* objectData,
                       int oid,
                       bool anyObject,
                       pool::DataCallBack& callback );

      void* loadPoolPointer( ObjectRelationalReadDataBuffer& buffer,
                             void* objectData,
                             int oid,
                             bool anyObject,
                             pool::DataCallBack& callback );

      bool eraseLoader( const IPtrLoader* loader );

      private:
      
      bool readObjectData( coral::ICursor& queryData,
                           ObjectRelationalReadDataBuffer& buffer,
                           void* objectData,
                           int& oid,
                           pool::DataCallBack& callback );

      /// Reads an array
      bool readArray( const std::vector<int>& identity,
                      ObjectRelationalReadDataBuffer& buffer,
                      void* address,
                      pool::DataCallBack& callback );

      /// Reads a pool ptr
      bool readPoolPointer( ObjectRelationalReadDataBuffer& buffer,
                            void* objectData,
                            int oid,
                            bool anyObject,
                            pool::DataCallBack& callback );
    private:

      coral::IQuery* makeQuery( ObjectRelationalReadDataBuffer& buffer,
                                int oid,
                                bool anyObject);


      bool processDependentBuffers( ObjectRelationalReadDataBuffer& buffer,
                                    void* objectData,
                                    int oid,
                                    bool anyObject,
                                    pool::DataCallBack& callback );      
      
    private:
      /// The map of mapping elements to element entries
      ObjectRelationalReadDataBuffer* m_dataBuffer;
      /// The dictionary for a pool reference
      const TypeH m_refDict;
      std::set<pool::IPtrLoader*> m_poolPointerLoaders;
    };

  }

}

#endif
